import asyncio
import concurrent.futures
import itertools
import time
from dataclasses import dataclass
from io import BytesIO
from typing import AsyncGenerator, Generator, Tuple

import httpx
import requests
from PIL import Image

async_client = httpx.AsyncClient()


DEFAULT_MAX_RETRIES = 6


@dataclass
class TileInfo:
    x: int
    y: int
    fileurl: str


@dataclass
class Tile:
    x: int
    y: int
    image: Image.Image


def get_width_and_height_from_zoom(zoom: int) -> Tuple[int, int]:
    """
    Returns the width and height of a panorama at a given zoom level, depends on the
    zoom level.
    """
    return 2**zoom, 2 ** (zoom - 1)


def make_download_url(pano_id: str, zoom: int, x: int, y: int) -> str:
    """
    Returns the URL to download a tile.
    """
    return (
        "https://cbk0.google.com/cbk"
        f"?output=tile&panoid={pano_id}&zoom={zoom}&x={x}&y={y}"
    )


def fetch_panorama_tile(
    tile_info: TileInfo, max_retries: int = DEFAULT_MAX_RETRIES
) -> Image.Image:
    """
    Tries to download a tile, returns a PIL Image.
    """
    for _ in range(max_retries):
        try:
            response = requests.get(tile_info.fileurl, stream=True)
            return Image.open(BytesIO(response.content))
        except requests.ConnectionError:
            print("Connection error. Trying again in 2 seconds.")
            time.sleep(2)
    raise requests.ConnectionError("Max retries exceeded.")


async def fetch_panorama_tile_async(
    tile_info: TileInfo, max_retries: int = DEFAULT_MAX_RETRIES
) -> Image.Image:
    """
    Asynchronously tries to download a tile, returns a PIL Image.
    """
    for _ in range(max_retries):
        try:
            response = await async_client.get(tile_info.fileurl)
            return Image.open(BytesIO(response.content))

        except httpx.RequestError as e:
            print(f"Request error {e}. Trying again in 2 seconds.")
            await asyncio.sleep(2)

    raise httpx.RequestError("Max retries exceeded.")


def iter_tile_info(pano_id: str, zoom: int) -> Generator[TileInfo, None, None]:
    """
    Generate a list of a panorama's tiles and their position.
    """
    width, height = get_width_and_height_from_zoom(zoom)
    for x, y in itertools.product(range(width), range(height)):
        yield TileInfo(
            x=x,
            y=y,
            fileurl=make_download_url(pano_id=pano_id, zoom=zoom, x=x, y=y),
        )


def iter_tiles(
    pano_id: str,
    zoom: int,
    max_retries: int = DEFAULT_MAX_RETRIES,
    multi_threaded: bool = False,
) -> Generator[Tile, None, None]:
    if not multi_threaded:
        for info in iter_tile_info(pano_id, zoom):
            image = fetch_panorama_tile(info, max_retries)
            yield Tile(x=info.x, y=info.y, image=image)
        return

    with concurrent.futures.ThreadPoolExecutor() as executor:
        future_to_tile = {
            executor.submit(fetch_panorama_tile, info, max_retries): info
            for info in iter_tile_info(pano_id, zoom)
        }
        for future in concurrent.futures.as_completed(future_to_tile):
            info = future_to_tile[future]
            try:
                image = future.result()
            except Exception as exc:
                raise Exception(
                    f"Failed to download tile {info.fileurl} due to Exception: {exc}"
                )
            else:
                yield Tile(x=info.x, y=info.y, image=image)


async def iter_tiles_async(
    pano_id: str, zoom: int, max_retries: int = DEFAULT_MAX_RETRIES
) -> AsyncGenerator[Tile, None]:
    for info in iter_tile_info(pano_id, zoom):
        image = await fetch_panorama_tile_async(info, max_retries)
        yield Tile(x=info.x, y=info.y, image=image)
    return


def get_panorama(
    pano_id: str,
    zoom: int = 5,
    multi_threaded: bool = False,
    max_retries: int = DEFAULT_MAX_RETRIES,
) -> Image.Image:
    """
    Downloads a streetview panorama.
    Multi-threaded is a lot faster, but it's also a lot more likely to get you banned.
    """
    tile_width = 512
    tile_height = 512

    total_width, total_height = get_width_and_height_from_zoom(zoom)
    panorama = Image.new("RGB", (total_width * tile_width, total_height * tile_height))

    for tile in iter_tiles(
        pano_id=pano_id,
        zoom=zoom,
        multi_threaded=multi_threaded,
        max_retries=max_retries,
    ):
        panorama.paste(im=tile.image, box=(tile.x * tile_width, tile.y * tile_height))
        del tile

    return panorama


async def get_panorama_async(
    pano_id: str, zoom: int, max_retries: int = DEFAULT_MAX_RETRIES
) -> Image.Image:
    """
    Downloads a streetview panorama by iterating through the tiles asynchronously.
    This runs in about the same speed as `get_panorama` with `multi_threaded=True`.
    """
    tile_width = 512
    tile_height = 512

    total_width, total_height = get_width_and_height_from_zoom(zoom)
    panorama = Image.new("RGB", (total_width * tile_width, total_height * tile_height))

    async for tile in iter_tiles_async(
        pano_id=pano_id, zoom=zoom, max_retries=max_retries
    ):
        panorama.paste(im=tile.image, box=(tile.x * tile_width, tile.y * tile_height))
        del tile

    return panorama



from PIL import Image


def get_pixel_from_bw_image(img: Image.Image, x: int, y: int) -> float:
    """
    Get the pixel value from a black and white image.
    """
    bw_img = img.convert("L")
    pixel = bw_img.getpixel((x, y))
    if pixel is None or isinstance(pixel, tuple):
        raise ValueError("Invalid pixel value")
    return pixel


def crop_bottom_and_right_black_border(img: Image.Image) -> Image.Image:
    """
    Crop the black border at the bottom and right of the panorama.
    When you download streetviews get_panorama(), it's common to see
    black borders at the bottom and right of the image. This is because
    he dimensions of the panorama are not always correct / multiple of
    512, a common issue with user-contributed panoramas.

    The implementation is not perfect, but it works for most cases.
    """
    (width, height) = img.size
    bw_img = img.convert("L")
    black_luminance = 4

    # Find the bottom of the panorama
    pixel_cursor = (0, height - 1)
    valid_max_y = height - 1
    while pixel_cursor[0] < width and pixel_cursor[1] >= 0:
        pixel_color = get_pixel_from_bw_image(bw_img, *pixel_cursor)

        if pixel_color > black_luminance:
            # Found a non-black pixel
            # Double check if all the pixels below this one are black
            all_pixels_below = list(
                bw_img.crop((0, pixel_cursor[1] + 1, width, height)).getdata()
            )
            all_black = True
            for pixel in all_pixels_below:
                if pixel > black_luminance:
                    all_black = False

            if all_black:
                valid_max_y = pixel_cursor[1]
                break
            else:
                # A false positive, probably the actual valid bottom pixel is
                # very close to black Reset the cursor to the next vertical
                # line to the right
                pixel_cursor = (pixel_cursor[0] + 1, height - 1)

        else:
            pixel_cursor = (pixel_cursor[0], pixel_cursor[1] - 1)

    # Find the right of the panorama
    pixel_cursor = (width - 1, 0)
    valid_max_x = width - 1
    while pixel_cursor[1] < height and pixel_cursor[0] >= 0:
        pixel_color = get_pixel_from_bw_image(bw_img, *pixel_cursor)

        if pixel_color > black_luminance:
            # Found a non-black pixel
            # Double check if all the pixels to the right of this one are black
            all_pixels_to_the_right = list(
                bw_img.crop((pixel_cursor[0] + 1, 0, width, height)).getdata()
            )
            all_black = True
            for pixel in all_pixels_to_the_right:
                if pixel > black_luminance:
                    all_black = False
            if all_black:
                valid_max_x = pixel_cursor[0]
                break
            else:
                # A false positive, probably the actual valid right pixel
                # is very close to black Reset the cursor to the next
                # horizontal line below
                pixel_cursor = (width - 1, pixel_cursor[1] + 1)

        else:
            pixel_cursor = (pixel_cursor[0] - 1, pixel_cursor[1])

    valid_height = valid_max_y + 1
    valid_width = valid_max_x + 1

    if valid_height == height and valid_width == width:
        # No black border found
        return img

    print(
        "Found black border."
        f" Cropping from {width}x{height} to {valid_width}x{valid_height}"
    )
    return img.crop((0, 0, valid_width, valid_height))



from io import BytesIO
from typing import Dict, Union

import requests
from PIL import Image
from pydantic import BaseModel


class Location(BaseModel):
    lat: float
    lng: float


class MetaData(BaseModel):
    date: str
    location: Location
    pano_id: str
    copyright: str


def get_panorama_meta(pano_id: str, api_key: str) -> MetaData:
    """
    Returns a panorama's metadata.

    Quota: This function doesn't use up any quota or charge on your API_KEY.

    Endpoint documented at:
    https://developers.google.com/maps/documentation/streetview/metadata
    """
    url = (
        "https://maps.googleapis.com/maps/api/streetview/metadata"
        f"?pano={pano_id}&key={api_key}"
    )
    resp = requests.get(url)
    return MetaData(**resp.json())


def get_streetview(
    pano_id: str,
    api_key: str,
    width: int = 640,
    height: int = 640,
    heading: int = 0,
    fov: int = 120,
    pitch: int = 0,
) -> Image.Image:
    """
    Get an image using the official API. These are not panoramas.

    You can find instructions to obtain an API key here:
    https://developers.google.com/maps/documentation/streetview/

    Args:
        pano_id (str): The panorama id.
        heading (int): The heading of the photo. Each photo is taken with a 360
            camera. You need to specify a direction in degrees as the photo
            will only cover a partial region of the panorama. The recommended
            headings to use are 0, 90, 180, or 270.
        api_key (str): Your API key.
        width (int): Image width (max 640 for non-premium downloads).
        height (int): Image height (max 640 for non-premium downloads).
        fov (int): Image field-of-view.
        pitch (int): Image pitch.
    """

    url = "https://maps.googleapis.com/maps/api/streetview"
    params: Dict[str, Union[str, int]] = {
        "size": "%dx%d" % (width, height),
        "fov": fov,
        "pitch": pitch,
        "heading": heading,
        "pano": pano_id,
        "key": api_key,
    }

    response = requests.get(url, params=params, stream=True)
    img = Image.open(BytesIO(response.content))
    return img



import json
import re
from typing import List, Optional

import requests
from pydantic import BaseModel
from requests.models import Response


class Panorama(BaseModel):
    pano_id: str
    lat: float
    lon: float
    heading: float
    pitch: Optional[float]
    roll: Optional[float]
    date: Optional[str]
    elevation: Optional[float]


def make_search_url(lat: float, lon: float) -> str:
    """
    Builds the URL of the script on Google's servers that returns the closest
    panoramas (ids) to a give GPS coordinate.
    """
    url = (
        "https://maps.googleapis.com/maps/api/js/"
        "GeoPhotoService.SingleImageSearch"
        "?pb=!1m5!1sapiv3!5sUS!11m2!1m1!1b0!2m4!1m2!3d{0:}!4d{1:}!2d50!3m10"
        "!2m2!1sen!2sGB!9m1!1e2!11m4!1m3!1e2!2b1!3e2!4m10!1e1!1e2!1e3!1e4"
        "!1e8!1e6!5m1!1e2!6m1!1e2"
        "&callback=callbackfunc"
    )
    return url.format(lat, lon)


def search_request(lat: float, lon: float) -> Response:
    """
    Gets the response of the script on Google's servers that returns the
    closest panoramas (ids) to a give GPS coordinate.
    """
    url = make_search_url(lat, lon)
    return requests.get(url)


def extract_panoramas(text: str) -> List[Panorama]:
    """
    Given a valid response from the panoids endpoint, return a list of all the
    panoids.
    """

    # The response is actually javascript code. It's a function with a single
    # input which is a huge deeply nested array of items.
    blob = re.findall(r"callbackfunc\( (.*) \)$", text)[0]
    data = json.loads(blob)

    if data == [[5, "generic", "Search returned no images."]]:
        return []

    subset = data[1][5][0]

    raw_panos = subset[3][0]

    if len(subset) < 9 or subset[8] is None:
        raw_dates = []
    else:
        raw_dates = subset[8]

    # For some reason, dates do not include a date for each panorama.
    # the n dates match the last n panos. Here we flip the arrays
    # so that the 0th pano aligns with the 0th date.
    raw_panos = raw_panos[::-1]
    raw_dates = raw_dates[::-1]

    dates = [f"{d[1][0]}-{d[1][1]:02d}" for d in raw_dates]

    return [
        Panorama(
            pano_id=pano[0][1],
            lat=pano[2][0][2],
            lon=pano[2][0][3],
            heading=pano[2][2][0],
            pitch=pano[2][2][1] if len(pano[2][2]) >= 2 else None,
            roll=pano[2][2][2] if len(pano[2][2]) >= 3 else None,
            date=dates[i] if i < len(dates) else None,
            elevation=pano[3][0] if len(pano) >= 4 else None,
        )
        for i, pano in enumerate(raw_panos)
    ]


def search_panoramas(lat: float, lon: float) -> List[Panorama]:
    """
    Gets the closest panoramas (ids) to the GPS coordinates.
    """

    resp = search_request(lat, lon)
    pans = extract_panoramas(resp.text)
    return pans


def parse_url(url: str) -> tuple[str, ...]:
    """
    extracts the lat, lon and pano_id from url
    """
    matches = re.search(
        r"\/@([-+]?[0-9]+[.]?[0-9]*),([-+]?[0-9]+[.]?[0-9]*).+!1s(.+)!2e", url
    )

    if matches:
        return matches.groups()

    return ("", "", "")


def search_panoramas_url(url: str) -> list[Panorama]:
    """
    Gets the closest panoramas (ids) to the GPS coordinates in the url.
    """
    lat, lon, _ = parse_url(url)
    return search_panoramas(float(lat), float(lon))


def search_panoramas_url_exact(url: str) -> Panorama | None:
    """
    Searches for exact panorama in url
    """
    _, _, id = parse_url(url)

    panos = search_panoramas_url(url)
    panos = [pano for pano in panos if pano.pano_id == id]

    return panos[0] if len(panos) > 0 else None



from .api import get_panorama_meta, get_streetview  # noqa
from .download import get_panorama, get_panorama_async  # noqa
from .search import (  # noqa
    search_panoramas,
    search_panoramas_url,
    search_panoramas_url_exact,
)
from .tools import crop_bottom_and_right_black_border  # noqa
